#                           PROTIS 2021/2022
#                            Samuel GERENTE
#
# Institut d'Optique Graduate School (https://www.institutoptique.fr/)
# LEnsE (http://lense.institutoptique.fr/)
#
# This python file implement the Dijkstra’s algorithm in order to find the shortest path in a maze-type grid.
# This work is hugely inspired by a project by Max Reynolds available here (https://github.com/maxwellreynolds/Maze).

import numpy as np

## Utilisation:

"""" 1) Définir une carte à l'aide d'une matrice de type array (importer numpy)
# maze = np.array([[1, 0, 1, 0, 1, 1, 1, 1, 1, 0],
#                 [1, 1, 1, 0, 0, 1, 0, 0, 1, 0],
#                 [1, 0, 1, 0, 0, 0, 1, 0, 1, 1],
#                 [0, 1, 1, 1, 1, 0, 1, 0, 0, 1],
#                 [0, 1, 0, 0, 1, 1, 1, 0, 1, 1],
#                 [1, 1, 1, 0, 1, 0, 1, 1, 1, 0],
#                 [1, 0, 0, 1, 0, 0, 0, 0, 1, 0],
#                 [1, 1, 1, 1, 0, 1, 1, 1, 1, 0],
#                 [0, 0, 1, 0, 0, 1, 0, 0, 1, 0],
#                 [0, 1, 1, 1, 1, 1, 0, 1, 1, 0]])*255
#
# # # 2) Chercher les coordonées des cases traversées par le meilleur chemin  et les stocker dans une liste
# pathe = find_shortest_path(maze, (0, 0), (0, 4))


# # # Pour directement visualiser le résultat exécuter la séquence suivante
# chemin = np.copy(maze)
# for case in pathe:
#     chemin[case] = 127
# plt.matshow(maze)
# plt.matshow(chemin)
# plt.show()
"""

## Fonctions

class Square:
    """
    Class for representing a square in the grid
    """
    def __init__(self, c, r):
        self.col = c  # column of the square
        self.row = r  # row of the square
        self.d = float('inf')  # distance from starting square
        self.parent_col = None  # column of the parent of the square
        self.parent_row = None  # row of the parent of the square
        self.processed = False  # is the square has already been process in the Dijkstra’s algorithm
        self.index_in_queue = None  # index in the processing min-heap queue of the Dijkstra’s algorithm

def get_neighbors(mat, r, c):
    """
    Return neighbor directly above, below, right, and left the square in pos (r, c) in mat.
    mat is a matrix containing object of type Square defined by the class above
    """
    shape = mat.shape
    neighbors = []
    # ensure neighbors are within image boundaries
    if r > 0 and not mat[r - 1, c].processed:  # above neighbor
        neighbors.append(mat[r - 1, c])
    if r < shape[0] - 1 and not mat[r + 1, c].processed:  # below neighbor
        neighbors.append(mat[r + 1, c])
    if c > 0 and not mat[r, c - 1].processed:  # left neighbor
        neighbors.append(mat[r, c - 1])
    if c < shape[1] - 1 and not mat[r, c + 1].processed:  # right neighbor
        neighbors.append(mat[r, c + 1])
    return neighbors


def bubble_up(queue, index):
    """
    Swap the given node (which index in the queue is index) with his parent until the queue is a min-heap
    """
    if index <= 0:
        return queue
    p_index = (index - 1) // 2  # index of the parent of the starting node
    if queue[index].d < queue[p_index].d:
        queue[index], queue[p_index] = queue[p_index], queue[index]
        queue[index].index_in_queue = index
        queue[p_index].index_in_queue = p_index
        queue = bubble_up(queue, p_index)
    return queue


def bubble_down(queue, index):
    """
    Swap the given node (which index in the queue is index) with his children until the queue is a min-heap
    """
    length = len(queue)
    lc_index = 2 * index + 1  # index of the left child of the starting node
    rc_index = lc_index + 1  # index of the right child of the starting node

    if lc_index >= length:  # no child
        return queue
    if lc_index < length <= rc_index:  # just left child
        if queue[index].d > queue[lc_index].d:
            queue[index], queue[lc_index] = queue[lc_index], queue[index]
            queue[index].index_in_queue = index
            queue[lc_index].index_in_queue = lc_index
            queue = bubble_down(queue, lc_index)
    else:  # left and right child
        small = lc_index
        if queue[lc_index].d > queue[rc_index].d:
            small = rc_index
        if queue[index].d > queue[small].d:
            queue[index], queue[small] = queue[small], queue[index]
            queue[index].index_in_queue = index
            queue[small].index_in_queue = small
            queue = bubble_down(queue, small)
    return queue


def get_distance(grid, r0, c0, r1, c1):
    """
    Return the distance between the square at (r0, c0) and the square at (r1, c1) in grid
    """
    return 0.1 + (grid[r0, c0] - grid[r1, c1])**2


def find_shortest_path(grid, start, end):
    """
    Find the shortest path between the square start and the square end computed using the Dijkstra’s algorithm.
    start and end are tuples containing the row and the column of the square : start = (row, column)
    grid is a matrix which contain only 2 number : 0 and 255. 0 (resp. 255) correspond to a square on which we can
    (resp. cannot) pass.
    """
    pq = []  # min-heap priority queue
    start_row = start[0]
    start_col = start[1]
    end_row = end[0]
    end_col = end[1]
    n_rows, n_cols = grid.shape[0], grid.shape[1]

    matrix = np.full((n_rows, n_cols), None)  # matrix that will contain all the squares of the grid
    for r in range(n_rows):
        for c in range(n_cols):
            matrix[r, c] = Square(c, r)
            matrix[r, c].index_in_queue = len(pq)
            pq.append(matrix[r, c])

    matrix[start_row, start_col].d = 0  # the distance's attribute of the starting square is set to 0

    pq = bubble_up(pq, matrix[start_row, start_col].index_in_queue)  # pq list is ordered into a min-heap queue
    # the starting square is the root (distance = 0) and all other nodes have infinite distance

    while len(pq) > 0:  # we iterate until all square have been processed

        # The square corresponding to the root of pq is process (because his distance is minimal)
        u = pq[0]
        u.processed = True
        # Then he is remove from pq (and the list in reorder to maintain the min-heap queue priority)
        pq[0] = pq[-1]
        pq[0].index_in_queue = 0
        pq.pop()
        pq = bubble_down(pq, 0)

        neighbors = get_neighbors(matrix, u.row, u.col)  # unprocessed neighbors are gathered
        for v in neighbors:
            # for each neighbor the distance is calculated and the min-heap is updated if necessary
            dist = get_distance(grid, u.row, u.col, v.row, v.col)
            if u.d + dist < v.d:
                v.d = u.d + dist
                v.parent_row = u.row
                v.parent_col = u.col
                idx = v.index_in_queue
                pq = bubble_down(pq, idx)
                pq = bubble_up(pq, idx)

    # the shortest path is then found
    path = []
    iter_v = matrix[end_row, end_col]
    path.append((end_row, end_col))
    while iter_v.row != start_row or iter_v.col != start_col:
        path.append((iter_v.row, iter_v.col))
        iter_v = matrix[iter_v.parent_row][iter_v.parent_col]
    path.append((start_row, start_col))

    return path